/*********************************************************************************/
/* Matrix product program with MPI on a virtual ring of processors               */
/* S. Vialle - December 2019                                                     */
/*********************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <mpi.h>
#include <omp.h>

//on Kyle :
#include <cblas.h>
//on Sarah :
//#include <openblas/cblas.h>

#include "main.h"
#include "kernel.h"


/*-------------------------------------------------------------------------------*/
/* Sequential product of local matrixes (optimized seq product).                 */
/*-------------------------------------------------------------------------------*/
void OneLocalProduct(unsigned long step)
{
  unsigned long OffsetStepLineC;

  // Set the number of OpenMP threads inside parallel regions
  omp_set_num_threads(NbThreads);

  // Compute the current step offset, in the MPI program, to access right C lines
  OffsetStepLineC = ((Me+step)%NbPE)*LOCAL_SIZE;

  switch (KernelId) {

  // Kernel 0 : Optimized code implemented by an application developer
  case 0 :
    #pragma omp parallel for
    for (int i = 0; i < LOCAL_SIZE; i++) {
      unsigned long iAoff = i*SIZE;
      unsigned long iCoff = (i+OffsetStepLineC)*LOCAL_SIZE;
      for (int k = 0; k < SIZE; k++) {
        unsigned long kBoff = k*LOCAL_SIZE;
        for (int j = 0; j < LOCAL_SIZE; j++) {
          C_Slice[iCoff+j] += A_Slice[iAoff+k] * B_Slice[kBoff+j];
        }
      }
    }
    break;

  // Kernel 1 : Very optimized computing kernel implemented in a HPC library
  case 1 :
    #pragma omp parallel
    {
      unsigned long reste = LOCAL_SIZE % omp_get_num_threads();
      unsigned long quotient = LOCAL_SIZE / omp_get_num_threads();
      unsigned long NbLineTh = quotient + 
                    (omp_get_thread_num() < reste ? 1 : 0);
      unsigned long offsetThLineC = quotient*omp_get_thread_num() + 
                         (omp_get_thread_num() < reste ? omp_get_thread_num() : reste);
      cblas_dgemm(CblasRowMajor, CblasNoTrans, CblasNoTrans,
                  NbLineTh, LOCAL_SIZE, SIZE,
                  1.0, &A_Slice[offsetThLineC*SIZE], SIZE, 
                  B_Slice, LOCAL_SIZE,
                  0.0, &C_Slice[(OffsetStepLineC+offsetThLineC)*LOCAL_SIZE], LOCAL_SIZE);
    }
    break;

  default :
    fprintf(stderr,"Error: kernel %d not implemented!\n",KernelId);
    exit(EXIT_FAILURE);
    break;
  }
}
